{Create graphic matching between items in two list boxex. Useful when conversion
of data from one set to another to perform visual matching.

Two listboxes are exposured as properties, height and left(right) alignments take
place authomatically when the properties are assigned. List boxex must NOT have
style LBS_SORTED. DragNotify event occur when one connects two items in different
listboxes. If returned value will be true then two items can be connected an back.

Run-time property MatchCount informs about amount of lines in the leftbox. Array
Match (base 0, maximal index MatchCount-1) for each line in left listbox gives
matched line in the right listbox. Method AutoListMatch perfomes authomatical
matching, it is base upon:
  a. Case-unsensitive comparison of lines
  b. Identity of objects for lines


Listboxes can be settled at run-time for an object, that exposes listbos (THeaderControl).

FreeWare. No warranties. If one will distribute the source code, reference on
author (below) is required. If one will modify the source code, highlighting
(by any means) of modifications is required.

Author: Sergei Trepalin
E-mail: trep@trep.home.chg.ru}


unit SetMatch;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls;

type
  TDragNotifyEvent=procedure(LeftPoint,RightPoint:integer; var CanDrop:boolean) of object;

  TSetsMatch = class(TCustomControl)
  private
    { Private declarations }
    FLeftParentCoor,FRightParentCoor,CoupledPairs:TList;
    FLastHighlight:integer;
    FDragNotify:TDragNotifyEvent;
    FDragPoint,FStartDragPoint:TPoint;
    FDragging:boolean;
    FCanDrop:boolean;
    FPartnerStoreNumber:integer;
    FLeftList,FRightList:TListBox;

    OldWndProcLeft,OldWndProcRight: TFarProc;
    NewWndProcLeft,NewWndProcRight: Pointer;

    { Utility functions }
    procedure HookLeft;
    procedure UnhookLeft;
    procedure HookRight;
    procedure UnhookRight;
    procedure HookWndProcLeft(var Message: TMessage);
    procedure HookWndProcRight(var Message: TMessage);
    procedure SetLeft(Value:TListBox);
    procedure SetRight(Value:TListBox);

    procedure InsertItem(BeforeIndex:integer; IsLeft:boolean);
    procedure DeleteItem(ItemNo:integer; IsLeft:boolean);
    procedure DrawRect(Canvas:TCanvas; Rect:TRect);
    procedure ProcessHighlight(MousePoint:TPoint);
    procedure RecalculateCoordinates(LBox:TListBox; Coor:TList);
    function  GetMatchCount:integer;
    function  GetMatch(index:cardinal):integer;
    procedure SetMatch(index:cardinal; Value:integer);
    procedure WMMouseMove(var Message:TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMLButtonDown(var Message:TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Message:TWMLButtonUp); message WM_LBUTTONUP;
  protected
    { Protected declarations }
    procedure Paint; override;
    function GetYCenter(ItemNo:integer; PCoor:TList):integer;
    function GetHighlightCoor(HighlightNumber:integer):TPoint; virtual;
    function GetPartner(HighlightNumber:integer):integer; virtual;
    function ItemFromPoint(MousePoint:TPoint):integer; virtual;
    procedure InvertBox(ItemNo:integer; LBox:TListBox); virtual;
  public
    { Public declarations }
    constructor Create(AOwner:TComponent); override;
    destructor  Destroy; override;
    procedure   AutoListMatch; virtual;
    property    MatchCount:integer read GetMatchCount;
    property    Match[index:cardinal]:integer read GetMatch write SetMatch; default;
  published
    { Published declarations }
    property DragNotify:TDragNotifyEvent read FDragNotify write FDragNotify;
    property LeftList:TListBox read FLeftList write SetLeft;
    property RightList:TListBox read FRightList write SetRight;
    property Color;
  end;

procedure Register;

implementation

const
  DefItemWidth=5;

{-----------------------------------------------------------------------------}
constructor TSetsMatch.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  Color:=clBtnFace;
  FLeftParentCoor:=TList.Create;
  FRightParentCoor:=TList.Create;
  CoupledPairs:=TList.Create;
  FDragging:=False;
  FCanDrop:=False;
end;

destructor TSetsMatch.Destroy;
begin
  if Assigned(FLeftParentCoor) then FLeftParentCoor.Free;
  if Assigned(FRightParentCoor) then FRightParentCoor.Free;
  if Assigned(CoupledPairs) then CoupledPairs.Free;
  try
    UnhookLeft;
    UnhookRight;
  except
  end;
  inherited Destroy;
end;

procedure TSetsMatch.HookLeft;
begin
  if FLeftList=nil then exit;
  OldWndProcLeft:=TFarProc(GetWindowLong(FLeftList.Handle, GWL_WNDPROC));
  NewWndProcLeft:=MakeObjectInstance(HookWndProcLeft);
  SetWindowLong(FLeftList.Handle,GWL_WNDPROC,LongInt(NewWndProcLeft));
end;

procedure TSetsMatch.UnhookLeft;
begin
  if (FLeftList<>nil) and IsWindow(FLeftList.Handle) and Assigned(OldWndProcLeft) then
    SetWindowLong(FLeftList.Handle,GWL_WNDPROC,longint(OldWndProcLeft));
  if Assigned(NewWndProcLeft) then FreeObjectInstance(NewWndProcLeft);
  NewWndProcLeft:=nil;
  OldWndProcLeft:=nil;
  FLeftList:=nil;
end;

procedure TSetsMatch.HookWndProcLeft(var Message:TMessage);
begin
  if FLeftList=nil then exit;
  with Message do begin
    if (Msg=WM_VSCROLL) then begin
      Result:=CallWindowProc(OldWndProcLeft,FLeftList.Handle,Msg,wParam,lParam);
      RecalculateCoordinates(FLeftList,FLeftParentCoor);
    end else if (Msg=LB_ADDSTRING) or (Msg=LB_RESETCONTENT) then begin
      Result:=CallWindowProc(OldWndProcLeft,FLeftList.Handle,Msg,wParam,lParam);
      RecalculateCoordinates(FLeftList,FLeftParentCoor);
    end else if Msg=LB_DELETESTRING then begin
      Result:=CallWindowProc(OldWndProcLeft,FLeftList.Handle,Msg,wParam,lParam);
      DeleteItem(Message.wParam,True);
      RecalculateCoordinates(FLeftList,FLeftParentCoor);
    end else if Msg=LB_INSERTSTRING then begin
      Result:=CallWindowProc(OldWndProcLeft,FLeftList.Handle,Msg,wParam,lParam);
      InsertItem(Message.wParam,True);
      RecalculateCoordinates(FLeftList,FLeftParentCoor);
    end else if Msg=WM_DESTROY then begin
      Result:=CallWindowProc(OldWndProcLeft,FLeftList.Handle,Msg,wParam,lParam);
      UnhookLeft;
    end else if Msg=WM_PAINT then begin
      Result:=CallWindowProc(OldWndProcLeft,FLeftList.Handle,Msg,wParam,lParam);
//      if FLastHighlight>0 then InvertBox(FLastHighlight-1,FLeftList);
    end else Result:=CallWindowProc(OldWndProcLeft,FLeftList.Handle,Msg,wParam,lParam);
  end;
end;

procedure TSetsMatch.SetLeft(Value:TListBox);
var
  WC:TWinControl;
begin
  UnhookLeft;
  FLeftList:=Value;
  HookLeft;
  if Assigned(FLeftList) then begin
    WC:=FLeftList;
    while (WC.Parent<>Parent) and (WC<>nil) do WC:=WC.Parent;
    if WC<>nil then begin
      WC.Left:=Left-FLeftList.Width;
      WC.Top:=Top;
      WC.Height:=Height;
    end;
  end;
  RecalculateCoordinates(FLeftList,FLeftParentCoor);
end;

procedure TSetsMatch.HookRight;
begin
  if FRightList=nil then exit;
  OldWndProcRight:=TFarProc(GetWindowLong(FRightList.Handle, GWL_WNDPROC));
  NewWndProcRight:=MakeObjectInstance(HookWndProcRight);
  SetWindowLong(FRightList.Handle,GWL_WNDPROC,LongInt(NewWndProcRight));
end;

procedure TSetsMatch.UnhookRight;
begin
  if (FRightList<>nil) and IsWindow(FRightList.Handle) and Assigned(OldWndProcRight) then
    SetWindowLong(FRightList.Handle,GWL_WNDPROC,longint(OldWndProcRight));
  if Assigned(NewWndProcRight) then FreeObjectInstance(NewWndProcRight);
  NewWndProcRight:=nil;
  OldWndProcRight:=nil;
  FRightList:=nil;
end;

procedure TSetsMatch.HookWndProcRight(var Message:TMessage);
begin
  if FRightList=nil then exit;
  with Message do begin
    if (Msg=WM_VSCROLL) then begin
      Result:=CallWindowProc(OldWndProcRight,FRightList.Handle,Msg,wParam,lParam);
      RecalculateCoordinates(FRightList,FRightParentCoor);
    end else if (Msg=LB_ADDSTRING) or (Msg=LB_RESETCONTENT) then begin
      Result:=CallWindowProc(OldWndProcRight,FRightList.Handle,Msg,wParam,lParam);
      RecalculateCoordinates(FRightList,FRightParentCoor);
    end else if Msg=LB_DELETESTRING then begin
      Result:=CallWindowProc(OldWndProcRight,FRightList.Handle,Msg,wParam,lParam);
      DeleteItem(Message.wParam,False);
      RecalculateCoordinates(FRightList,FRightParentCoor);
    end else if Msg=LB_INSERTSTRING then begin
      Result:=CallWindowProc(OldWndProcRight,FRightList.Handle,Msg,wParam,lParam);
      InsertItem(Message.wParam,False);
      RecalculateCoordinates(FRightList,FRightParentCoor);
    end else if Msg=WM_DESTROY then begin
      Result:=CallWindowProc(OldWndProcRight,FRightList.Handle,Msg,wParam,lParam);
      UnhookRight;
    end else if Msg=WM_PAINT then begin
      Result:=CallWindowProc(OldWndProcRight,FRightList.Handle,Msg,wParam,lParam);
//      if FLastHighlight>0 then InvertBox(FLastHighlight-1,FRightList);
    end else Result:=CallWindowProc(OldWndProcRight,FRightList.Handle,Msg,wParam,lParam);
  end;
end;

procedure TSetsMatch.SetRight(Value: TListBox);
var
  WC:TWinControl;
begin
  UnhookRight;
  FRightList:=Value;
  HookRight;
  if Assigned(FRightList) then begin
    WC:=FRightList;
    while (WC.Parent<>Parent) and (WC<>nil) do WC:=WC.Parent;
    if WC<>nil then begin
      WC.Left:=Left+Width;
      WC.Top:=Top;
      WC.Height:=Height;
    end;
  end;
  RecalculateCoordinates(FRightList,FRightParentCoor);
end;


procedure TSetsMatch.DrawRect(Canvas:TCanvas; Rect:TRect);
begin
  if Canvas.Handle=0 then Exit;
  Canvas.MoveTo(Rect.Left,Rect.Top);
  Canvas.LineTo(Rect.Right,Rect.Top);
  Canvas.LineTo(Rect.Right,Rect.Bottom);
  Canvas.LineTo(Rect.Left,Rect.Bottom);
  Canvas.LineTo(Rect.Left,Rect.Top);
end;

function TSetsMatch.GetHighlightCoor(HighlightNumber:integer):TPoint;
{Returns Coor of Highlight point {+HighlightNumber - left, -HighlightNumber-right}
begin
  Result.X:=-32500; Result.Y:=-32500;
  if HighlightNumber>0 then begin
    Result.X:=5+DefItemWidth;
    Result.Y:=GetYCenter(HighlightNumber,FLeftParentCoor);
  end else if HighlightNumber<0 then begin
    Result.X:=ClientWidth-DefItemWidth-5;
    Result.Y:=GetYCenter(abs(HighlightNumber),FRightParentCoor);
  end;
end;

function TSetsMatch.GetPartner(HighlightNumber:integer):integer;
{Returns assigned partner for given highlight number}
var
  I:integer;
  L:longint;
  W1,W2:word;
begin
  Result:=0;
  FPartnerStoreNumber:=-1;
  if (not Assigned(CoupledPairs)) or (CoupledPairs.Count=0) or (HighlightNumber=0) then Exit;
  for I:=0 to CoupledPairs.Count-1 do begin
    L:=longint(CoupledPairs[I]);
    W1:=LoWord(L);
    W2:=HiWord(L);
    if HighlightNumber>0 then begin {Right point search}
      if W1=HighlightNumber then begin
        Result:=-W2;
        FPartnerStoreNumber:=I;
        Exit;
      end;
    end else if HighlightNumber<0 then begin
      if W2=abs(HighlightNumber) then begin
        Result:=W1;
        FPartnerStoreNumber:=I;
        Exit;
      end;
    end;
  end;
end;

function TSetsMatch.ItemFromPoint(MousePoint:TPoint):integer;
{Returns highlight number from point}
var
  I,Y:integer;
begin
  Result:=0;
  if (MousePoint.X>=5) and (MousePoint.X<=5+2*DefItemWidth) then begin
    if FLeftParentCoor.Count>1 then for I:=1 to FLeftParentCoor.Count-1 do begin
      Y:=GetYCenter(I,FLeftParentCoor);
      if (MousePoint.Y>=(Y-DefItemWidth)) and (MousePoint.Y<=(Y+DefItemWidth)) then begin
        Result:=I;
        Exit;
      end;
    end;
  end else if (MousePoint.X<=(ClientWidth-5)) and (MousePoint.X>=(ClientWidth-5-2*DefItemWidth)) then begin
    if FRightParentCoor.Count>1 then for I:=1 to FRightParentCoor.Count-1 do begin
      Y:=GetYCenter(I,FRightParentCoor);
      if (MousePoint.Y>=(Y-DefItemWidth)) and (MousePoint.Y<=(Y+DefItemWidth)) then begin
        Result:=-I;
        Exit;
      end;
    end;
  end;
end;

procedure TSetsMatch.ProcessHighlight(MousePoint:TPoint);
var
  LastHighlight:integer;
  DC:HDC;
  PH:TPoint;
begin
  LastHighlight:=ItemFromPoint(MousePoint);
  if LastHighlight<>FLastHighlight then begin
    DC:=0;
    try
      if Canvas.Handle=0 then begin
        DC:=GetDC(Handle);
        Canvas.Handle:=DC;
      end;
      Canvas.Pen.Color:=clBlack;
      Canvas.Pen.Mode:=pmCopy;
      if FLastHighlight<>0 then begin
        Canvas.Brush.Color:=clWhite;
        PH:=GetHighlightCoor(FLastHighlight);
        Canvas.Rectangle(PH.X-DefItemWidth,PH.Y-DefItemWidth,PH.X+DefItemWidth,PH.Y+DefItemWidth);
        if FLastHighlight>0 then InvertBox(FLastHighlight-1,FLeftList)
        else InvertBox(abs(FLastHighlight)-1,FRightList);
      end;
      if LastHighlight<>0 then begin
        Canvas.Brush.Color:=clAqua;
        PH:=GetHighlightCoor(LastHighlight);
        Canvas.Rectangle(PH.X-DefItemWidth,PH.Y-DefItemWidth,PH.X+DefItemWidth,PH.Y+DefItemWidth);
        if LastHighlight>0 then InvertBox(LastHighlight-1,FLeftList)
        else InvertBox(abs(LastHighlight)-1,FRightList);
      end;
      if DC<>0 then begin
        Canvas.Handle:=0;
        ReleaseDC(Handle,DC);
        DC:=0;
      end;
    except
      if DC<>0 then begin
        Canvas.Handle:=0;
        ReleaseDC(Handle,DC);
      end;
    end;
    FLastHighlight:=LastHighlight;
  end;
end;

procedure TSetsMatch.WMMouseMove(var Message:TWMMouseMove);
var
  DC:HDC;
  PH:TPoint;
  Highlight:integer;
begin
  inherited;
  PH.X:=Message.XPos;
  PH.Y:=Message.YPos;
  ProcessHighlight(PH);
  if (vk_Left and Message.Keys)>0 then begin
    DC:=0;
    if FDragging then try
      if Canvas.Handle=0 then begin
        DC:=GetDC(Handle);
        Canvas.Handle:=DC;
      end;
      Canvas.Pen.Color:=Color;
      Canvas.Pen.Mode:=pmXOR;
      Canvas.MoveTo(FStartDragPoint.X,FStartDragPoint.Y);
      Canvas.LineTo(FDragPoint.X,FDragPoint.Y);
      Canvas.MoveTo(FStartDragPoint.X,FStartDragPoint.Y);
      Canvas.LineTo(Message.XPos,Message.YPos);
      FDragPoint.X:=Message.XPos;
      FDragPoint.Y:=Message.YPos;
      if DC<>0 then begin
        Canvas.Handle:=0;
        ReleaseDC(Handle,DC);
      end;
      if FLastHighlight<>0 then begin
        Highlight:=ItemFromPoint(FStartDragPoint);
        if Highlight*FLastHighlight>0 then FCanDrop:=False else begin
          FCanDrop:=True;
          if Assigned(FDragNotify) then begin
            {External counter 0..Count-1}
            if Highlight<0 then FDragNotify(FLastHighlight-1,abs(Highlight)-1,FCanDrop)
            else FDragNotify(Highlight-1,abs(FLastHighlight)-1,FCanDrop);
          end;
        end;
      end else FCanDrop:=True;
      if FCanDrop then Screen.Cursor:=crDrag else Screen.Cursor:=crNoDrop;
    except
      if DC<>0 then begin
        Canvas.Handle:=0;
        ReleaseDC(Handle,DC);
      end;
    end;
  end;
  Message.Result:=0;
end;

procedure TSetsMatch.WMLButtonDown(var Message:TWMLButtonDown);
var
  PH:TPoint;
  HPartner:integer;
begin
  inherited;
  if (not FDragging) and (FLastHighlight<>0) then begin
    HPartner:=GetPartner(FLastHighlight);
    PH:=GetHighlightCoor(FLastHighlight);
    PH:=ClientToScreen(PH);
    SetCursorPos(PH.X,PH.Y);
    FDragPoint:=ScreenToClient(PH);
    FDragging:=True;
    if HPartner<>0 then FStartDragPoint:=GetHighlightCoor(HPartner)
    else FStartDragPoint:=FDragPoint;
    FCanDrop:=True;
    Cursor:=crDrag;
  end;
  Message.Result:=0;
end;

procedure TSetsMatch.WMLButtonUp(var Message:TWMLButtonUp);
var
  N:integer;
  W1,W2:word;
  L:longint;
begin
  inherited;
  if FDragging then begin
    FDragging:=False;
    Canvas.Pen.Mode:=pmCopy;
    Screen.Cursor:=crDefault;
    N:=ItemFromPoint(FStartDragPoint);
    if FLastHighlight=0 then begin  {Remove assignment if necessary}
      if N<>0 then begin
        GetPartner(N);
        if FPartnerStoreNumber>=0 then CoupledPairs.Delete(FPartnerStoreNumber);
      end;
    end else if FCanDrop then begin {Create assignment}
      {Clear assignment for last highlight}
      GetPartner(FLastHighlight);
      if FPartnerStoreNumber>=0 then CoupledPairs.Delete(FPartnerStoreNumber);
      {Clear assignment for initial point}
      GetPartner(N);
      if FPartnerStoreNumber>=0 then CoupledPairs.Delete(FPartnerStoreNumber);
      if FLastHighlight<0 then begin
        W1:=N;
        W2:=abs(FLastHighlight);
      end else begin
        W1:=FLastHighlight;
        W2:=abs(N);
      end;
      L:=MakeLong(W1,W2);
      CoupledPairs.Add(pointer(L));
    end;
    CoupledPairs.Pack;
    Cursor:=crDefault;
    Invalidate;
  end;
  Message.Result:=0;
end;

function TSetsMatch.GetYCenter(ItemNo:integer; PCoor:TList):integer;
var
  P:TPoint;
begin
  Result:=-1;
  if (ItemNo<1) or (ItemNo>(PCoor.Count-1)) then Exit;
  if Parent=nil then Exit;
  P.X:=0;
  P.Y:=integer(PCoor[ItemNo-1])+(integer(PCoor[ItemNo])-integer(PCoor[ItemNo-1])) div 2;
  P:=TWinControl(Parent).ClientToScreen(P);
  P:=ScreenToClient(P);
  Result:=P.Y;
end;

procedure TSetsMatch.Paint;
var
  I:integer;
  L:longint;
  W1,W2:word;
  PH:TPoint;
begin
  if Canvas.Handle=0 then Exit;
  Canvas.Brush.Color:=clWhite;
  Canvas.Pen.Color:=clBlack;
  Canvas.Pen.Mode:=pmCopy;
  if CoupledPairs.Count>0 then for I:=0 to CoupledPairs.Count-1 do begin
    L:=longint(CoupledPairs[I]);
    W1:=LoWord(L);
    W2:=HiWord(L);
    PH:=GetHighlightCoor(W1);
    Canvas.MoveTo(PH.X,PH.Y);
    PH:=GetHighlightCoor(-W2);
    Canvas.LineTo(PH.X,PH.Y);
  end;
  if FLeftParentCoor.Count>1 then for I:=1 to FLeftParentCoor.Count-1 do begin
    PH:=GetHighlightCoor(I);
    DrawRect(Canvas,Rect(PH.X-DefItemWidth,PH.Y-DefItemWidth,PH.X+DefItemWidth,PH.Y+DefItemWidth));
    Canvas.Rectangle(PH.X-DefItemWidth,PH.Y-DefItemWidth,PH.X+DefItemWidth,PH.Y+DefItemWidth);
  end;
  if FRightParentCoor.Count>1 then for I:=1 to FRightParentCoor.Count-1 do begin
    PH:=GetHighlightCoor(-I);
    DrawRect(Canvas,Rect(PH.X-DefItemWidth,PH.Y-DefItemWidth,PH.X+DefItemWidth,PH.Y+DefItemWidth));
    Canvas.Rectangle(PH.X-DefItemWidth,PH.Y-DefItemWidth,PH.X+DefItemWidth,PH.Y+DefItemWidth);
  end;
  if FLastHighlight<>0 then begin
    Canvas.Brush.Color:=clAqua;
    PH:=GetHighlightCoor(FLastHighlight);
    Canvas.Rectangle(PH.X-DefItemWidth,PH.Y-DefItemWidth,PH.X+DefItemWidth,PH.Y+DefItemWidth);
  end;
  Canvas.Brush.Color:=Color;
end;

procedure TSetsMatch.RecalculateCoordinates(LBox:TListBox; Coor:TList);
var
  I:integer;
  R:TRect;
  P:TPOint;
begin
  {Re-dimension array if necessary}
  if Coor.Count<>(LBox.Items.Count+1) then begin
    Coor.Clear;
    Coor.Pack;
    Coor.Capacity:=LBox.Items.Count+1;
    if LBox.Items.Count<>0 then for I:=0 to LBox.Items.Count do Coor.Add(pointer(1));
  end;
  if LBox.Items.Count>0 then begin
    R:=LBox.ItemRect(0);
    P.X:=0; P.Y:=R.Top;
    P:=LBox.ClientToScreen(P);
    P:=Parent.ScreenToClient(P);
    Coor[0]:=pointer(P.Y);
    for I:=0 to LBox.Items.Count-1 do begin
      R:=LBox.ItemRect(I);
      P.X:=0; P.Y:=R.Bottom;
      P:=LBox.ClientToScreen(P);
      P:=Parent.ScreenToClient(P);
      Coor[I+1]:=pointer(P.Y);
    end;
  end;
  Invalidate;
end;

procedure TSetsMatch.InvertBox(ItemNo:integer; LBox:TListBox);
var
  DC:HDC;
  R:TRect;
begin
  if not Assigned(LBox) then Exit;
  if (ItemNo<0) or (ItemNo>=LBox.Items.Count) then Exit;
  DC:=0;
  try
    if LBox.Canvas.Handle=0 then begin
      DC:=GetDC(LBox.Handle);
      LBox.Canvas.Handle:=DC;
    end;
    R:=LBox.ItemRect(ItemNo);
    InvertRect(LBox.Canvas.Handle,R);
    if DC<>0 then begin
      LBox.Canvas.Handle:=0;
      ReleaseDC(LBox.Handle,DC);
      DC:=0;
    end;
  except
    if DC<>0 then begin
      LBox.Canvas.Handle:=0;
      ReleaseDC(LBox.Handle,DC);
    end;
  end;
end;

procedure TSetsMatch.DeleteItem(ItemNo:integer; IsLeft:boolean);
var
  I:integer;
  L:longint;
  W1,W2:word;
begin
  Inc(ItemNo); {Matching - list box is started from zero index}
  if (not Assigned(CoupledPairs)) or (CoupledPairs.Count=0) then Exit;
  for I:=CoupledPairs.Count-1 downto 0 do begin
    L:=longint(CoupledPairs[I]);
    W1:=LoWord(L);
    W2:=HiWord(L);
    if (IsLeft and (W1=ItemNo)) or ((not IsLeft) and (W2=ItemNo)) then CoupledPairs.Delete(I);
  end;
  CoupledPairs.Pack;
  if CoupledPairs.Count>0 then for I:=0 to CoupledPairs.Count-1 do begin
    L:=longint(CoupledPairs[I]);
    W1:=LoWord(L);
    W2:=HiWord(L);
    if (IsLeft and (W1>ItemNo)) then W1:=W1-1 else
    if ((not IsLeft) and (W2>ItemNo)) then W2:=W2-1;
    L:=MakeLong(W1,W2);
    CoupledPairs[I]:=pointer(L);
  end;
end;

procedure TSetsMatch.InsertItem(BeforeIndex:integer; IsLeft:boolean);
var
  I:integer;
  L:longint;
  W1,W2:word;
begin
  if (not Assigned(CoupledPairs)) or (CoupledPairs.Count=0) then Exit;
  if CoupledPairs.Count>0 then for I:=0 to CoupledPairs.Count-1 do begin
    L:=longint(CoupledPairs[I]);
    W1:=LoWord(L);
    W2:=HiWord(L);
    if (IsLeft and (W1>BeforeIndex)) then W1:=W1+1 else
    if ((not IsLeft) and (W2>BeforeIndex)) then W2:=W2+1;
    L:=MakeLong(W1,W2);
    CoupledPairs[I]:=pointer(L);
  end;
end;

function TSetsMatch.GetMatchCount:integer;
begin
  Result:=0;
  if Assigned(FLeftList) then Result:=FLeftList.Items.Count;
end;

function TSetsMatch.GetMatch(index:cardinal):integer;
var
  I:integer;
  W1,W2:word;
  L:longint;
begin
  Result:=-1;
  if Assigned(CoupledPairs) and (CoupledPairs.Count>0) then for I:=0 to CoupledPairs.Count-1 do begin
    L:=longint(CoupledPairs[I]);
    W1:=LoWord(L);
    W2:=HiWord(L);
    if W1=(index+1) then begin
      Result:=W2-1;
      Exit;
    end;
  end;
end;

procedure TSetsMatch.SetMatch(index:cardinal; Value:integer);
var
  I,N1,N2:integer;
  L:longint;
  W1,W2:word;
begin
  if (not Assigned(CoupledPairs)) then Exit;
  if (not Assigned(FLeftList)) or (not Assigned(FRightList)) then Exit;
  N1:=Index+1;
  N2:=Value+1;
  if (N1<1) or (N1>FLeftList.Items.Count) or (N2<1) or (N2>FRightList.Items.Count) then Exit;
  if CoupledPairs.Count>0 then for I:=CoupledPairs.Count-1 downto 0 do begin
    L:=longint(CoupledPairs[I]);
    W1:=LoWord(L);
    W2:=HiWord(L);
    if (W1=N1) or (W2=N2) then CoupledPairs.Delete(I);
  end;
  CoupledPairs.Pack;
  W1:=N1;
  W2:=N2;
  L:=MakeLong(W1,W2);
  CoupledPairs.Add(pointer(L));
end;

procedure TSetsMatch.AutoListMatch;
var
  I,J:integer;
  List:TList;
  L:longint;
  W1,W2:word;
  S1,S2:string;
begin
  if (not Assigned(FLeftList)) or (not Assigned(FRightList)) then Exit;
  if (FLeftList.Items.Count=0) or (FRightList.Items.Count=0) then Exit;
  CoupledPairs.Clear;
  CoupledPairs.Pack;
  List:=nil;
  try
    List:=TList.Create;
    if Assigned(List) then begin
      List.Capacity:=FRightList.Items.Count;
      for I:=1 to FRightList.Items.Count do List.Add(nil);
      for I:=0 to FLeftList.Items.Count-1 do begin
        S1:=FLeftList.Items[I];
        for J:=0 to FRightList.Items.Count-1 do if List[J]=nil then begin
          S2:=FRightList.Items[J];
          if (FLeftList.Items.Objects[I]=FRightList.Items.Objects[J]) and (ANSICompareText(S1,S2)=0) then begin
            W1:=I+1;
            W2:=J+1;
            L:=MakeLong(W1,W2);
            CoupledPairs.Add(pointer(L));
            List[J]:=pointer(1);
          end;
          if List[J]<>nil then break;
        end;
      end;
      List.Free;
      List:=nil;
    end;
  except
    if Assigned(List) then List.Free;
  end;
end;
{-------------------------------------------------------------------------}
procedure Register;
begin
  RegisterComponents('Interface', [TSetsMatch]);
end;

end.
